// Original code by TAKAHASHI Shuhei <pcb@pcbsoft.net>
// This code is licensed under NYSL ver. 0.9982.
// See LICENSE.txt for details.

#define WIN32_LEAN_AND_MEAN
#include <winsock2.h>
#include <ws2tcpip.h>
#include <mstcpip.h>

#include <windows.h>

#include <rasdlg.h>
#include <ras.h>
#include <raserror.h>

#include <set>
#include <utility>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <ctype.h>

#include <assert.h>

#include "Buffer.h"
#include "Packet.h"
#include "hspdll.h"
#include "hspparam.h"

#include "md5.h"


/*
	HRTt@C쐬̂߂̃}N`
*/

#ifdef HRT
#undef EXPORT
#define EXPORT extern "C"
#endif // HRT


/*
	GCAXHSP߂쐬
*/

#define ALIAS(alias, org) \
	EXPORT BOOL WINAPI alias(int p1, int p2, int p3, int p4) { \
		return ( ( BOOL (WINAPI *) (int,int,int,int) ) org )(p1, p2, p3, p4); \
	}




/*
	낢
*/

#define INADDR_ERROR	INADDR_ANY

#define MAX_PACKET		1500
#define MAX_SOCKET		4			// ListenvgR
#define HOSTNAMESIZE	256
#define URLSIZE			256
#define MAX_XMLSIZE		16384
#define IGDDISCOVERY_TIMEOUT	3000
#define HTTP_TIMEOUT	3000
#define MAX_LINELEN		65536
#define DYNAMIC_PORT_BEGIN	59001
#define DYNAMIC_PORT_END	60000

#define ST_UNDEFINED	0
#define ST_TCP			1
#define ST_TCPLIS		2
#define ST_UDP			4
#define ST_RAW			8
#define ST_DNS			16
#define ST_PING			32
#define ST_SPECIAL		1024
#define ST_TCPALL		(ST_TCP|ST_TCPLIS)
#define ST_ALL			(~0)

#define ASYNC_DNS		1
#define ASYNC_NATMAP	2
#define ASYNC_PING		4

#define ICMP_ECHO		8
#define ICMP_REPLY		0
#define ICMP_UNREACHED	3
#define ICMP_TTL		11
#define ICMP_NOREPLY	255

#define WM_PCBNET_ASYNC				(WM_USER+100)
#define WM_PCBNET_DNSREQUEST		(WM_USER+102)
#define WM_PCBNET_DNSREPLY			(WM_USER+103)
#define WM_PCBNET_SENDING			(WM_USER+105)
#define WM_PCBNET_CONNECT			(WM_USER+106)
#define WM_PCBNET_BIND				(WM_USER+107)
#define WM_PCBNET_ACCEPT			(WM_USER+108)
#define WM_PCBNET_SOCKATTACH		(WM_USER+109)
#define WM_PCBNET_PING				(WM_USER+117)
#define WM_PCBNET_THREAD_ENDING		(WM_USER+118)
#define WM_PCBNET_TRY_CONNECT		(WM_USER+119)

#define NATSTAT_NOTINIT				0
#define NATSTAT_INITIALIZING		1
#define NATSTAT_AVAILABLE			2
#define NATSTAT_NOTAVAILABLE		3

#define NATERR_IGDNOTFOUND			1
#define NATERR_IGDNOTSUPP			2
#define NATERR_NETDOWN				3

#define NATERR_PORTNOTAVAIL			4
#define NATERR_FAILEDREGISTMAP		5
#define NATERR_FAILEDUNREGISTMAP	6



// \PbgEBhE

#define PCBNET_WNDCLASS		"pcbnet2_asyncwnd"
#define PCBNET_WNDTITLE		"pcbnet2_asyncwnd"



/*
	NXE\̒`
*/



// SOCKETPAIR\

typedef struct {
	SOCKET soc;
	struct sockaddr_storage addr;
	int addrlen;
} SOCKETPAIR;



// IGDINFO\

typedef struct {
	char DeviceURL[URLSIZE];
	char DeviceHost[URLSIZE];
	char DevicePath[URLSIZE];
	unsigned short DevicePort;
	char DeviceType[128];
	char ControlURL[URLSIZE];
	char ControlHost[URLSIZE];
	char ControlPath[URLSIZE];
	unsigned short ControlPort;
	char ExternalIPAddress[64];
} IGDINFO;


// HTTPHEADER\
typedef struct {
	char Tag[64];
	char Value[URLSIZE];
} HTTPHEADER;


// HTTPRESPONSE\
typedef struct {
	char Protocol[16];
	int Code;
	list<HTTPHEADER> Headers;
} HTTPRESPONSE;


// HTTPREQUEST\
typedef struct {
	char Path[256];
	list<HTTPHEADER> Headers;
	char *Body;
	HTTPRESPONSE res;
} HTTPREQUEST;


// PORTMAPPING\
typedef struct {
	int socktype;
	struct sockaddr_in inaddr;
	struct sockaddr_in exaddr;
	char description[256];
	BOOL bAvailable;
	int err;
} PORTMAPPING;


typedef struct {
	DWORD createTime;
	unsigned int sentOctets;
	unsigned int receivedOctets;
} STATISTICS;




// ip_hdr/icmp_hdr\

#pragma pack(4)

typedef struct {
	unsigned char headerlen:4;
	unsigned char ipver:4;
	unsigned char tos;
	unsigned short len;
	unsigned short id;
	unsigned short fr;
	unsigned char ttl;
	unsigned char proto;
	unsigned short checksum;
	unsigned long srcip;
	unsigned long destip;
} ip_hdr;

typedef struct {
	unsigned char Type;
	unsigned char SubCode;
	unsigned short CheckSum;
	union {
		unsigned long reserved;
		struct {
			union {
				struct {
					unsigned short Id;
					unsigned short Seq;
				};
				struct {
					unsigned long dwId;
				};
			};
		} ping;
	};
} icmp_hdr;


#pragma pack()


//	SOCKETINFONX

class SOCKETINFO {
private:
	CRITICAL_SECTION crit;

public:
	BOOL bInit;			// tO
	int type;			// \Pbg^Cv
	HWND hWnd;			// \PbgEBhEnh
	DWORD threadID;		// XbhID
	HANDLE hThread;		// Xbhnh
	int err;			// G[R[h
	DWORD dwUser;		// [U[`l
	int fAsyncTask;		// 񓯊^XNtO
	BOOL bClosing;		// \PbgN[YtO
	PORTMAPPING map;	// NAT|[g}bsO
	SOCKET soc[MAX_SOCKET];		// \PbgfBXNv^
	int socs;					// \Pbg
	unsigned short family[MAX_SOCKET];		// \Pbgt@~
	struct sockaddr_storage addr;	// MAhX
	int addrlen;				// addr
	STATISTICS stat;	// vf[^
	union {
		struct {
			char hostname[HOSTNAMESIZE];	// ڑzXg
			BOOL bSendable;				// M\tO
			BOOL bBlocking;				// ubLOtO
			BOOL bShutting;				// shutdowntO
			BOOL bSending;				// MtO
			BOOL bReadPending;			// recv()\tO
			int iBufferLen;				// obt@LpVeB
			Buffer *r;					// Mobt@
			Buffer *s;					// Mobt@
			addrinfo *res0;				// getaddrinfop
			addrinfo *res;				// getaddrinfop
			BOOL bLineCut;				// tcpgetlp
		} tcp;
		struct {
			BOOL bListening;			// ҋ@tO
			BOOL bIncoming[MAX_SOCKET];	// ڑvtO
		} tcplis;
		struct {
			char hostname[HOSTNAMESIZE];	// ڑzXg
			BOOL bBound;				// boundtO
			BOOL bBlocking;				// ubLOtO
			BOOL bResolved;				// ς݃tO
			BOOL bSendable;				// M\tO
			BOOL bReadPending;			// recv()\tO
			int iBufferLen;				// obt@LpVeB
			Packet *r;					// Mobt@
			Packet *s;					// Mobt@
			struct sockaddr_storage fromaddr;	// ŏIMAhX
			int fromaddrlen;			// fromaddr
		} udp;
		struct {
			BOOL bReply;				// ԐMtO
			addrinfo *res0;				// getaddrinfop
			char hostname[HOSTNAMESIZE];	// ڑzXg
		} dns;
		struct {
			char hostname[HOSTNAMESIZE];	// PingzXg
			int iTimeOut;				// ő҂
			unsigned char TTL;			// TTL
			BOOL bResult;				// ԐMtO
			unsigned char replycode;	// ԓR[h
			int iTripTime;				// 
			struct sockaddr_storage fromaddr;	// ԓAhX
			int fromaddrlen;			// fromaddr
		} ping;
	};

	SOCKETINFO();
	~SOCKETINFO();
	void Lock();
	void Unlock();
};

typedef set<SOCKETINFO *> SOCKETSET;

typedef pair<unsigned char, unsigned short> PORT;
typedef set<PORT> PORTSET;




/*
	O[o֐錾
*/


BOOL Initialize();
BOOL Cleanup();
unsigned short CheckSum(char *buf, int len);

HWND CreateSocketWindow(SOCKETINFO *si);
SOCKETINFO * CreateSocket(int socktype);
BOOL IsSocket(SOCKETINFO *si, int type=ST_ALL);
BOOL IsSocketX(SOCKETINFO *si, int type=ST_ALL);
BOOL DeleteSocket(SOCKETINFO *si);
void DestroySocketSet();
void EndThread(SOCKETINFO *si);

void InitSocketManager();
void CleanupSocketManager();

LRESULT CALLBACK NetProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

int AddPortMapping(PORTMAPPING *mapping);
int DeletePortMapping(PORTMAPPING *mapping);
void AddPortMappingProc(void *param);
void DeletePortMappingProc(void *param);

char * getstr(char *dest, char *src, int destsize, char token = '\0');
char * nextline(char *p);
char * skipspace(char *p);
char * deletebackspace(char *p);

int ParseHTTPResponse(HTTPRESPONSE *res, char *in);
void RenewXML(char *xml);
BOOL GetXMLEntry(char *dest, int destsize, char *xml, char *tag);
int GetXMLEntryInt(char *xml, char *tag);
int SOAPAction(char *hostname, unsigned short port, char *devicetype, char *action, HTTPREQUEST *req, char *out, int outsize);


/*
	O[oϐ錾
*/

extern BOOL bInit;				// tO
extern HINSTANCE ghInstance;	// DLLHINSTANCE
extern int iTCPSendBuf;			// TCP̑MP
extern int iTCPRecvBuf;			// TCP̎MP
extern BOOL bStrict;			// Strict[hH
extern int iMaxBuffer;			// \Pbg̍őobt@
extern unsigned int iMaxUdpDg;	// UDPpPbg̍ő咷
extern WSADATA wsaData;			// WSADATA returned by WSAStartup()

extern SOCKETSET ss;
extern SOCKETSET ssx;



/*
	HSPɕԂG[R[h
*/

#define HSPERROR_INVALID_SOCKET		3




/*
	[N`FbNpR[h
*/

#ifdef _DEBUG
#include <crtdbg.h>
void memLeakCheck();
#define new new(_NORMAL_BLOCK,__FILE__,__LINE__)
#define malloc(p1) _malloc_dbg(p1,_NORMAL_BLOCK,__FILE__,__LINE__)
#define MEMLEAK_CHECK() memLeakCheck()
#else
#define MEMLEAK_CHECK()
#endif	// _DEBUG



/*
	ʂɂǂłǂ
*/

#pragma function(memset, memcpy, strcmp, strcpy, strlen, strcat)




